/*******************************************************************************
*
*	TITLE:		ir_sensor.c 
*
*	VERSION:	0.1 (Beta)                           
*
*	DATE:		22-Feb-2008
*
*	AUTHOR:		R. Kevin Watson
*	   			kevinw@jpl.nasa.gov
*
*	COMMENTS:	This software provides an interface to a Vishay infrared 
*				sensor which can be used to decode commands sent from a 
*				Sony television remote control or one on the many universal
*				remote controls configured to communicate with a Sony 
*				device. It receives Vishay TSOP4840 (or similar) infrared 
*				sensor waveforms encoded with the very common Sony 12-bit
*				remote control protocol and outputs the 5-bit address and 
*				7-bit command.
*
*				This software uses timer 1.
*
*				This software assumes that your Vishay IR sensor is
*				attached to digital I/O pins 3, 4, 5 or 6. See included
*				Vishay connection diagram document for instructions.
*
*				This file best viewed with tabs set to four.
*
*				You are free to use this source code for any non-commercial
*				use. Please do not make copies of this source code, modified
*				or un-modified, publicly available on the internet or elsewhere
*				without permission. Thanks.
*
*				Copyright 2008 R. Kevin Watson. All rights are reserved.
*
********************************************************************************
*
*	CHANGE LOG:
*
*	DATE         REV  DESCRIPTION
*	-----------  ---  ----------------------------------------------------------
*	22-Feb-2008  0.1  RKW - Original code.
*
*******************************************************************************/
#include "ifi_frc.h"
#include "ir_sensor.h"

// don't read these directly
volatile unsigned int command_address = 0;
volatile unsigned char ir_update_count = 0;

Sony_IR_Data_Struct ir_data;


/*******************************************************************************
*
*	FUNCTION:		IR_Sensor_ISR()
*
*	PURPOSE:		Receives Vishay TSOP4840 (or similar) infrared sensor
*					waveforms encoded with the very common Sony 12-bit
*					remote control protocol and outputs the 5-bit address
*					and 7-bit command.	
*
*	CALLED FROM:
*
*	PARAMETERS:		Unsigned char with the least-significant bit equal
*					to the state of the interrupt pin.
*
*	RETURNS:		Nothing, but does update the global variables address,
*					command, and ir_update_flag when new data is received.
*
*	COMMENTS:		This software uses timer 1.
*
*******************************************************************************/
#ifdef USE_ALTERNATE_TMPDATA
#pragma tmpdata low_isr_tmpdata
#endif
void IR_Sensor_ISR(unsigned char pin_state)
{
	unsigned int ticks;
	unsigned char temp_tmr1l;
	static unsigned int bit_weight = 0;
	static unsigned char synchronized = FALSE;
	static unsigned int temp_command_address;

	// falling edge of the sensor waveform?
	if(pin_state == 0)
	{
		// initialize the timer register
		TMR1H = 0x00;
		TMR1L = 0x00;

		// start the timer
		T1CONbits.TMR1ON = 1;
	}
	// rising edge of the sensor waveform
	else
	{
		// stop the timer
		T1CONbits.TMR1ON = 0;
	
		// convert the 8-bit values from the timer registers 
		// into a 16-bit value we can use in our calculations 
		temp_tmr1l = TMR1L;
		ticks = (unsigned int)TMR1H;
		ticks <<= 8;
		ticks += (unsigned int)temp_tmr1l;

		// in the context of this software, when Synchronized is true
		// we've received a valid 2.4ms start pulse and we're building
		// the 12-bit command and address value.
		if(synchronized == TRUE)
		{
			// did we receive a logic zero?
			if(ticks > LOGIC_ZERO_TICKS_LOW && ticks < LOGIC_ZERO_TICKS_HIGH)
			{
				// because we reset command_address to zero earlier,
				// we don't need to do anything other than wait for 
				// the next bit to arrive...
				bit_weight <<= 1;
			}
			// did we receive a logic one?
			else if(ticks > LOGIC_ONE_TICKS_LOW && ticks < LOGIC_ONE_TICKS_HIGH)
			{
				temp_command_address += bit_weight;
				bit_weight <<= 1;
			}
			else
			{
				// we received corrupted data; start over
				bit_weight = 1;
				temp_command_address = 0;
				synchronized = FALSE;	
			}
		}
		// waiting for a start bit
		else
		{
			// did we just receive a start bit?
			if(ticks > START_BIT_TICKS_LOW && ticks < START_BIT_TICKS_HIGH)
			{
				synchronized = TRUE;
			}
		}

		// if the thirteenth bit of bit_weight is set, we've
		// received a valid command and address data set
		if(bit_weight == 0x1000)
		{
			// update the global copy of command_address
			// with the fresh data re just received
			command_address = temp_command_address;

			// signal that new data is available
			ir_update_count++;

			// initialize the variable in preparation for a new data word
			bit_weight = 1;
			temp_command_address = 0;
			synchronized = FALSE;
		}
	}
}
#ifdef USE_ALTERNATE_TMPDATA
#pragma tmpdata
#endif

/*******************************************************************************
*
*	FUNCTION:		Initialize_IR_Sensor()
*
*	PURPOSE:		Initializes the IR sensor code.		
*
*	CALLED FROM:	Teleop.c/Initialization()
*
*	PARAMETERS:		None
*
*	RETURNS:		Nothing
*
*	COMMENTS:
*
*******************************************************************************/
void Initialize_IR_Sensor(void)
{
	// leave timer off for now
	T1CONbits.TMR1ON = 0;

	// use 1:1 prescaler (clock=10MHz/each tick=100ns)
	T1CONbits.T1CKPS0 = 0;
	T1CONbits.T1CKPS1 = 0;

	// timer 1 oscillator disabled
	T1CONbits.T1OSCEN = 0;

	// use the internal clock
	T1CONbits.TMR1CS = 0;

	// timer 1 register operations are done in one 16-bit access
	T1CONbits.RD16 = 1;

	// we don't need to use interrupts
	IPR1bits.TMR1IP = 0;
	PIR1bits.TMR1IF = 0;
	PIE1bits.TMR1IE = 0;

	// initialize the timer register
	TMR1H = 0x00;
	TMR1L = 0x00;
}

/*******************************************************************************
*
*	FUNCTION:		Get_IR_Sensor_Data()
*
*	PURPOSE:		Fills in the user's Sony_IR_Data_Struct	data structure
*					with the latest IR sensor data. 		
*
*	CALLED FROM:	User's code.
*
*	PARAMETERS:		Pointer to a Sony_IR_Data_Struct, which is defined
*					in ir_sensor.h
*
*	RETURNS:		Nothing
*
*	COMMENTS:
*
*******************************************************************************/
void Get_IR_Sensor_Data(Sony_IR_Data_Ptr ir_data)
{
	unsigned int temp_command_address;

	// briefly turn off interrupts to prevent data corruption
	INTCONbits.RBIE = 0;

	// get a local copy of the latest IR sensor data
	temp_command_address = command_address;

	// turn the interrupts back on
	INTCONbits.RBIE = 1;

	// fill the user's data structure
	ir_data->command = (unsigned char)(temp_command_address & 0x7F);
	temp_command_address >>= 7;
	ir_data->address = (unsigned char)(temp_command_address & 0x1F);
}

/*******************************************************************************
*
*	FUNCTION:		Get_IR_Sensor_Count()
*
*	PURPOSE:		Call this function to see if new data is available.		
*
*	CALLED FROM:	User's code.
*
*	PARAMETERS:		None
*
*	RETURNS:		Unsigned char containing the number of IR data sets
*					received since the last time Reset_IR_Sensor_Count()
*					was called.
*
*	COMMENTS:
*
*******************************************************************************/
unsigned char Get_IR_Sensor_Count(void)
{
	unsigned int temp_ir_update_count;

	// briefly turn off interrupts to prevent data corruption
	INTCONbits.RBIE = 0;

	// get a local copy of the latest IR update count
	temp_ir_update_count = ir_update_count;

	// turn the interrupts back on
	INTCONbits.RBIE = 1;

	return(temp_ir_update_count);
}

/*******************************************************************************
*
*	FUNCTION:		Reset_IR_Sensor_Count()
*
*	PURPOSE:		Resets the IR data sets counter to zero.		
*
*	CALLED FROM:	User's code.
*
*	PARAMETERS:		None
*
*	RETURNS:		Nothing
*
*	COMMENTS:
*
*******************************************************************************/
void Reset_IR_Sensor_Count(void)
{
	// briefly turn off interrupts to prevent data corruption
	INTCONbits.RBIE = 0;

	// reset the count
	ir_update_count = 0;

	// turn the interrupts back on
	INTCONbits.RBIE = 1;
}
